#include "higher_order_func.h"
#include <stdlib.h>
#include <string.h>

void *map(void *array, size_t n, size_t size, void (*fn)(void *elem)) {
  if (n == 0) {
    return NULL;
  }
  void *result = malloc(n * size);
  // copy original array to result byte by byte
  memcpy((char *)result, (char *)array, n * size);
  for (size_t i = 0; i < n; i++) {
    // cast it to char* to make it byte addressable
    fn((char *)result + i * size);
  }
  return result;
}

void *filter(void *array, size_t n, size_t size, size_t *new_size,
             bool (*pred)(void *elem)) {
  *new_size = 0;
  size_t count = 0;

  if (n == 0) {
    return NULL;
  }

  for (size_t i = 0; i < n; i++) {
    void *element = (char *)array + i * size;
    if (pred(element)) {
      count++;
    }
  }
  if (count == 0) {
    return NULL;
  }
  void *result = malloc(count * size);
  count = 0;
  for (size_t i = 0; i < n; i++) {
    void *element = (char *)array + i * size;
    if (pred(element)) {
      void *result_element = (char *)result + count * size;
      memcpy(result_element, element, size);
      count++;
    }
  }
  *new_size = count;
  return result;
}

void *reduce(void *array, size_t n, size_t size, void *initial_value,
             void *(*reducer)(void *acc, void *elem)) {
  if (n == 0) {
    return initial_value;
  }
  void *accumulator = initial_value;
  for (size_t i = 0; i < n; i++) {
    void *element = (char *)array + i * size;
    accumulator = reducer(accumulator, element);
  }
  return accumulator;
}
